#!/usr/bin/env python
# -*- coding: utf-8 -*-

# geoarray, A fast Python interface for image geodata - either on disk or in memory.
#
# Copyright (C) 2019  Daniel Scheffler (GFZ Potsdam, daniel.scheffler@gfz-potsdam.de)
#
# This software was developed within the context of the GeoMultiSens project funded
# by the German Federal Ministry of Education and Research
# (project grant code: 01 IS 14 010 A-C).
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
test_geoarray
-------------

Tests for the geoarray.baseclasses module.
"""


from unittest import TestCase
import tempfile
import os
from collections import OrderedDict
from typing import Iterable

import numpy as np
from shapely.geometry import Polygon
from osgeo.osr import SpatialReference
import matplotlib
from parameterized import parameterized
from py_tools_ds.geo.vector import geometry
from py_tools_ds.geo.coord_trafo import transform_any_prj

from geoarray import GeoArray, masks, __file__ as geoarray_root
from geoarray.metadata import GDAL_Metadata


__author__ = ["Daniel Scheffler", "Jessica Palka"]


# Path of the tests-directory in the geoarray-package.
tests_path = os.path.abspath(os.path.join(geoarray_root, "..", ".."))

# disable matplotlib figure popups
matplotlib.use('Template')  # disables matplotlib figure popups


path_data = os.path.join(tests_path, "tests", "data", "L8_2bands_extract10x11.tif")
gA_inMem = GeoArray(path_data)
gA_inMem.to_mem()
gA_inMem.filePath = ""

gA_notInMem = GeoArray(path_data)
params_inMem_NotInMem = [("inMem", gA_inMem),
                         ("notInMem", gA_notInMem)]


class Test_GeoArray(TestCase):
    # expected results
    given_bandnames = ['B1', 'B2']

    @classmethod
    def setUpClass(cls) -> None:
        cls.gA = gA_inMem

    @parameterized.expand(params_inMem_NotInMem)
    def test_bandnames(self, _, gA):
        """Test GeoArray.bandnames."""
        self.assertEqual(gA.bandnames, OrderedDict([('B1', 0), ('B2', 1)]))

    @parameterized.expand(params_inMem_NotInMem)
    def test_shape(self, _, gA):
        """Test GeoArray.shape, .ndim, .rows, .columns, and .bands"""
        shape_attrs = (gA.ndim, gA.rows, gA.columns, gA.bands)
        shape_property = ('DIMENSIONS', 'ROWS', 'COLUMNS', 'BANDS')
        R_exp, C_exp, B_exp = 10, 11, 2
        expected_result = (3, R_exp, C_exp, B_exp)  # dimensions, rows, columns, band

        self.assertEqual(gA.shape, (R_exp, C_exp, B_exp))

        for i in range(4):
            with self.subTest(i=i):
                self.assertEqual(shape_attrs[i], expected_result[i],
                                 msg='The number of {i} is different from the expected result!'.format(
                                     i=shape_property[i]))

    @parameterized.expand(params_inMem_NotInMem)
    def test_dtype(self, _, gA):
        """Test GeoArray.dtype."""
        self.assertEqual(gA.dtype, np.dtype('float32'))

    @parameterized.expand(params_inMem_NotInMem)
    def test_geotransform_gsd_grid(self, _, gA):
        """Test GeoArray.geotransform, .xgsd, .ygsd, and .xygrid_specs."""
        self.assertEqual(gA.geotransform, [365985.0, 30.0, 0.0, 5916615.0, 0.0, -30.0])
        self.assertEqual((gA.xgsd, gA.ygsd), (30, 30))
        self.assertEqual(gA.xygrid_specs, [[365985.0, 366015.0], [5916615.0, 5916585.0]],
                         msg='The [[xOrigin, xGSD], [yOrigin, yGSD]]-grid is not as expected!')

    @parameterized.expand(params_inMem_NotInMem)
    def test_projection_epsg(self, _, gA):
        """Test GeoArray.projection and .epsg."""

        # Convert WKT-string of the projection to a Proj4_string
        srs = SpatialReference()
        srs.ImportFromWkt(gA.projection)
        proj4 = srs.ExportToProj4()

        self.assertEqual(proj4.strip(' /t/n/r'), '+proj=utm +zone=33 +datum=WGS84 +units=m +no_defs')
        self.assertEqual(gA.epsg, 32633)

    def test_nodata_not_given(self):
        """Test GeoArray.nodata in case no nodata value is given (i.e., auto-detected)."""
        gA = GeoArray(np.random.randint(1, 10, (5, 10)))
        self.assertIsNone(gA._nodata)
        self.assertEqual(self.gA.nodata, -9999.0)

    def test_nodata_given(self):
        """Test GeoArray.nodata in case nodata value is given."""
        gA = GeoArray(np.array([1, 2]), nodata=-9999)
        self.assertEqual(gA.nodata, -9999.0)

    @parameterized.expand(params_inMem_NotInMem)
    def test_calc_mask_nodata(self, _, gA):
        def test_with_flag_all(_gA):
            """Test mask output with flag='all' (flag when nodata in ALL bands)."""
            mnd_all = _gA.calc_mask_nodata(overwrite=True, flag='all')
            mnd_all_exp = np.ones((10, 11)).astype(bool)
            mnd_all_exp[0, 0] = False  # False represents nodata
            self.assertTrue(np.array_equal(mnd_all, mnd_all_exp))

        def test_with_flag_any(_gA):
            """Test mask output with flag='any' (flag when nodata in ANY band)."""
            mnd_any = _gA.calc_mask_nodata(overwrite=True, flag='any')
            mnd_any_exp = np.ones((10, 11)).astype(bool)
            mnd_any_exp[0, 0:3] = False  # False represents nodata
            self.assertTrue(np.array_equal(mnd_any, mnd_any_exp))

        if gA.is_inmem:
            _real_nodata = gA.nodata
            try:
                for nodata in [_real_nodata, np.nan]:
                    gA.nodata = nodata

                    gA[0, 0, :] = nodata  # set pixel [0, 0] to nodata in all bands
                    gA[0, 1, 0] = nodata  # set pixel [0, 1] to nodata in first band only
                    gA[0, 2, 1] = nodata  # set pixel [0, 2] to nodata in second band only

                    test_with_flag_all(gA)
                    test_with_flag_any(gA)

            finally:
                gA.nodata = _real_nodata

            with self.assertRaises(ValueError):
                gA.calc_mask_nodata(overwrite=True, flag='bad_string')

    @parameterized.expand(params_inMem_NotInMem)
    def test___getitem__(self, _, gA):
        def validate(array, exp_shape):
            self.assertIsInstance(array, np.ndarray)
            self.assertEqual(array.shape, exp_shape)

        R, C, B = gA.shape  # (10, 11, 2)

        # test row/col subset
        validate(gA[:1, :3, :], (1, 3, B))  # only one row is requested, given as a slice
        validate(gA[0, :3, :], (3, B))  # only one row is requested, given as an int
        validate(gA[2:5, :3], (3, 3, B))  # third dimension is not given
        validate(gA[2:5, :3, :], (3, 3, B))

        # test band subset
        validate(gA[:, :, 0:1], (R, C, 1))  # band slice  # returns 3D array
        validate(gA[:, :, 0], (R, C))  # band indexing  # returns 2D array
        validate(gA[1], (R, C))  # only band is given  # returns 2D
        validate(gA['B1'], (R, C))  # only bandname is given

        # test wrong inputs
        self.assertRaises(ValueError, gA.__getitem__, 'B01')

        # test full array  # NOTE: This sets self.gA.arr!
        validate(gA[:], (R, C, B))

        # TODO: add further tests

    def test___getitem__consistency(self):
        testarr = np.zeros((2, 2, 2))
        testarr[0, :, :] = [[11, 12], [13, 14]]
        testarr[1, :, :] = [[21, 22], [23, 24]]

        gA_inmem = GeoArray(testarr)
        inmem_res = gA_inmem[0, :, :]

        with tempfile.TemporaryDirectory() as tf:
            gA_inmem.save(os.path.join(tf, 'test.bsq'))

            gA_notinmem = GeoArray(os.path.join(tf, 'test.bsq'))
            notinmem_res = gA_notinmem[0, :, :]

        self.assertEqual(inmem_res.ndim, notinmem_res.ndim)
        self.assertEqual(inmem_res.shape, notinmem_res.shape)

    def test___getitem__consistency_3d_array_1_column(self):
        testarr = np.array([[1, 2], [3, 4]]).reshape((2, 1, 2))

        gA_inmem = GeoArray(testarr)
        inmem_res = gA_inmem[:]

        with tempfile.TemporaryDirectory() as tf:
            gA_inmem.save(os.path.join(tf, 'test.bsq'))

            gA_notinmem = GeoArray(os.path.join(tf, 'test.bsq'))
            notinmem_res = gA_notinmem[:]

        self.assertEqual(inmem_res.ndim, notinmem_res.ndim)
        self.assertEqual(inmem_res.shape, notinmem_res.shape)

    def test___getitem__consistency_2d_array(self):
        testarr = np.zeros((2, 2))
        testarr[:, :] = [[11, 12], [13, 14]]

        gA_inmem = GeoArray(testarr)
        inmem_res = gA_inmem[0, 0]

        with tempfile.TemporaryDirectory() as tf:
            gA_inmem.save(os.path.join(tf, 'test.bsq'))

            gA_notinmem = GeoArray(os.path.join(tf, 'test.bsq'))
            notinmem_res = gA_notinmem[0, 0]

        self.assertEqual(inmem_res.ndim, notinmem_res.ndim)
        self.assertEqual(inmem_res.shape, notinmem_res.shape)

    def test_numpy_array(self):
        arr = np.array(self.gA)
        self.assertIsInstance(arr, np.ndarray)
        self.assertEqual(arr.shape, self.gA.shape)

    def test_box(self):
        """Test GeoArray.box."""
        self.assertIsInstance(self.gA.box, geometry.boxObj)

    def test_mask_nodata(self):
        # TODO: Consider the dependency of mask_nodata on calc_mask_nodata().
        """Test GeoArray.mask_nodata."""
        self.assertIsInstance(self.gA.mask_nodata, masks.NoDataMask)

    def test_mask_baddata(self):
        """Test GeoArray.mask_baddata."""
        bdm = masks.BadDataMask(os.path.join(tests_path, "tests", "data", "L8_BadDataMask10x11.tif"))

        for i in range(2):
            with self.subTest(i=1):
                if i == 0:
                    # mask_baddata-function is "None", when it is not actively set
                    self.assertIsNone(self.gA.mask_baddata)
                if i == 1:
                    self.gA.mask_baddata = bdm
                    self.assertIsInstance(self.gA.mask_baddata, masks.BadDataMask)

    def test_footprint_poly(self):
        # TODO: Test the validation of the footprint_poly-function.
        # TODO: Consider the dependencies of the footprint_poly-function on mask_nodata, boxObj.
        """Test GeoArray.footprint_poly."""
        self.assertIsInstance(self.gA.footprint_poly, Polygon)

    def test_metadata(self):
        # TODO: Create a metadata-file for the tested TIFF-Image.
        # TODO: Test, if the metadata-function gives an output
        """Test GeoArray.metadata."""
        self.assertIsInstance(self.gA.metadata, GDAL_Metadata)

    def test_tiles(self):
        test_gAs = [self.gA,  # 3D
                    GeoArray(self.gA[:, :, 0], geotransform=self.gA.gt, projection=self.gA.prj)]  # 2D

        for gA in test_gAs:
            tiles = gA.tiles(tilesize=(50, 50))
            self.assertIsInstance(tiles, Iterable)

            for ((rS, rE), (cS, cE)), tile in tiles:
                self.assertTrue(np.array_equal(tile, gA[rS: rE + 1, cS: cE + 1], equal_nan=True))

    def test_get_subset_3D_geoarray(self):
        # test without resetting band names
        sub_gA = self.gA.get_subset(xslice=slice(2, 5), yslice=slice(None, 3), zslice=slice(1, 2))
        self.assertIsInstance(sub_gA, GeoArray)
        self.assertTrue(list(sub_gA.bandnames), list(self.gA.bandnames)[1])

        # test with providing only xslice
        sub_gA = self.gA.get_subset(xslice=slice(2, 5))
        self.assertIsInstance(sub_gA, GeoArray)

        # test with providing only yslice
        sub_gA = self.gA.get_subset(yslice=slice(None, 3))
        self.assertIsInstance(sub_gA, GeoArray)

        # test with zslice provided as list
        sub_gA = self.gA.get_subset(xslice=slice(2, 5), yslice=slice(None, 3), zslice=[0, 1])
        self.assertIsInstance(sub_gA, GeoArray)

        # test without providing zslice
        sub_gA = self.gA.get_subset(xslice=slice(2, 5), yslice=slice(None, 3))
        self.assertIsInstance(sub_gA, GeoArray)

        # test requesting only one column
        sub_gA = self.gA.get_subset(xslice=slice(0, 1), yslice=slice(None, 3))
        self.assertIsInstance(sub_gA, GeoArray)

        # test with resetting band names
        sub_gA = self.gA.get_subset(xslice=slice(2, 5), yslice=slice(None, 3), zslice=slice(1, 2),
                                    reset_bandnames=True)
        self.assertTrue(list(sub_gA.bandnames), ['B1'])

        # test arrays are equal
        sub_gA = self.gA.get_subset(xslice=slice(2, 5), yslice=slice(None, 3), zslice=slice(1, 2))
        sub_arr = self.gA[:3, 2:5, 1:2]
        self.assertTrue(np.array_equal(sub_gA[:], sub_arr, equal_nan=True))

        # test deepcopied arrays (modification of sub_gA.arr must not affect self.gA.arr)
        sub_gA[:2, :2] = 99
        self.assertTrue(np.array_equal(sub_gA[:2, :2], np.full((2, 2, 1), 99, self.gA.dtype)))
        self.assertNotEqual(np.mean(sub_arr[:2, :2]), 99)
        self.assertNotEqual(np.std(sub_arr[:2, :2]), 0)

        # test metadata
        self.assertEqual(sub_gA.meta.bands, 1)
        self.assertEqual(len(list(sub_gA.meta.band_meta.values())[0]), 1)
        self.assertEqual(len(list(sub_gA.bandnames.keys())), 1)
        self.assertNotEqual(sub_gA.gt, self.gA.gt)
        self.assertEqual(sub_gA.prj, self.gA.prj)

        # test not to return GeoArray
        out = self.gA.get_subset(xslice=slice(2, 5), yslice=slice(None, 3), zslice=slice(1, 2),
                                 return_GeoArray=False)

        self.assertIsInstance(out, tuple)
        self.assertTrue(len(out), 3)
        self.assertIsInstance(out[0], np.ndarray)
        self.assertIsInstance(out[1], tuple)
        self.assertIsInstance(out[2], str)

    def test_get_subset_2D_geoarray(self):
        gA_2D = GeoArray(self.gA[0])

        # test without resetting band names
        sub_gA = self.gA.get_subset(xslice=slice(2, 5), yslice=slice(None, 3))
        self.assertIsInstance(sub_gA, GeoArray)
        self.assertTrue(list(sub_gA.bandnames), list(self.gA.bandnames)[1])

        # test with providing only xslice
        sub_gA = gA_2D.get_subset(xslice=slice(2, 5))
        self.assertIsInstance(sub_gA, GeoArray)

        # test with providing only yslice
        sub_gA = gA_2D.get_subset(yslice=slice(None, 3))
        self.assertIsInstance(sub_gA, GeoArray)

        # test without providing zslice
        sub_gA = gA_2D.get_subset(xslice=slice(2, 5), yslice=slice(None, 3))
        self.assertIsInstance(sub_gA, GeoArray)

        # test requesting only one column
        sub_gA = self.gA.get_subset(xslice=slice(0, 1), yslice=slice(None, 3))
        self.assertIsInstance(sub_gA, GeoArray)

        # test with resetting band names
        sub_gA = gA_2D.get_subset(xslice=slice(2, 5), yslice=slice(None, 3), reset_bandnames=True)
        self.assertTrue(list(sub_gA.bandnames), ['B1'])

        # test arrays are equal
        sub_gA = gA_2D.get_subset(xslice=slice(2, 5), yslice=slice(None, 3))
        sub_gA_2D = gA_2D[:3, 2:5]
        self.assertTrue(np.array_equal(sub_gA[:], sub_gA_2D))

        # test deepcopied arrays (modification of sub_gA.arr must not affect self.gA.arr)
        sub_gA[:2, :2] = 99
        self.assertTrue(np.array_equal(sub_gA[:2, :2], np.full((2, 2), 99, gA_2D.dtype)))
        self.assertNotEqual(np.mean(sub_gA_2D[:2, :2]), 99)
        self.assertNotEqual(np.std(sub_gA_2D[:2, :2]), 0)

        # test metadata
        self.assertEqual(sub_gA.meta.bands, 1)
        self.assertEqual(len(list(sub_gA.meta.band_meta.values())[0]), 1)
        self.assertEqual(len(list(sub_gA.bandnames.keys())), 1)
        self.assertNotEqual(sub_gA.gt, gA_2D.gt)
        self.assertEqual(sub_gA.prj, gA_2D.prj)

        # test not to return GeoArray
        out = gA_2D.get_subset(xslice=slice(2, 5), yslice=slice(None, 3), return_GeoArray=False)

        # test with provided zslice
        with self.assertRaises(ValueError):
            gA_2D.get_subset(xslice=slice(2, 5), yslice=slice(None, 3), zslice=slice(1, 2))
        with self.assertRaises(ValueError):
            gA_2D.get_subset(xslice=slice(2, 5), yslice=slice(None, 3), zslice=[1, 2])

        self.assertIsInstance(out, tuple)
        self.assertTrue(len(out), 3)
        self.assertIsInstance(out[0], np.ndarray)
        self.assertIsInstance(out[1], tuple)
        self.assertIsInstance(out[2], str)

    @parameterized.expand(params_inMem_NotInMem)
    def test_save(self, _, gA):
        """Test GeoArray.save."""
        # Save as temporary GTiff
        L8_2bands_extract10x11_copy = os.path.join(tests_path, "tests", "data", "output",
                                                   "L8_BadDataMask10x11_copy.tif")
        gA.save(L8_2bands_extract10x11_copy, fmt="GTiff")
        gA_copy = GeoArray(L8_2bands_extract10x11_copy)

        self.assertTrue(os.path.exists(L8_2bands_extract10x11_copy))
        self.assertIsInstance(gA_copy, GeoArray)
        self.assertTrue(np.array_equal(gA[:], gA_copy[:], equal_nan=True))

    def test_PlottingFunctions(self):
        self.gA.show()
        # self.gA.show(interactive=True) # only works if test is started with ipython.

        self.gA.show_map()
        self.gA.show_histogram()

    def test_get_mapPos(self):
        # TODO: validate results (pixel dimensions, values, metadata subset, ...)
        xmin, xmax, ymin, ymax = self.gA.box.boundsMap
        xgsd = self.gA.xgsd
        ygsd = self.gA.ygsd
        arr_sub, gt, prj = self.gA.get_mapPos(mapBounds=(xmin + xgsd, ymin, xmax + xgsd, ymax - ygsd / 2),
                                              mapBounds_prj=self.gA.prj)
        self.assertIsInstance(arr_sub, np.ndarray)
        self.assertIsInstance(gt, (tuple, list))
        self.assertIsInstance(prj, str)

    def test_read_pointData__singleCoord_oneband(self):
        vals_z = self.gA.read_pointData(mapXY_points=np.array([[366015, 5916585]]),
                                        band=1)

        self.assertIsInstance(vals_z, self.gA.dtype)
        self.assertEqual(vals_z, 5766)

    def test_read_pointData__singleCoord_allbands(self):
        vals_z = self.gA.read_pointData(mapXY_points=np.array([[366015, 5916585]]))

        self.assertIsInstance(vals_z, np.ndarray)
        self.assertEqual(vals_z.size, self.gA.bands)
        self.assertTrue(np.array_equal(vals_z, np.array([[5352, 5766]])))

    def test_read_pointData__singleCoord_allbands_otherPrj(self):
        X, Y = transform_any_prj(self.gA.epsg, 4326, 366015.1, 5916584.9)
        vals_z = self.gA.read_pointData(mapXY_points=np.array([[X, Y]]),
                                        mapXY_points_prj=4326)

        self.assertIsInstance(vals_z, np.ndarray)
        self.assertEqual(vals_z.size, self.gA.bands)
        self.assertTrue(np.array_equal(vals_z, np.array([[5352, 5766]])))

    def test_read_pointData__multiCoords_allbands(self):
        vals_z = self.gA.read_pointData(mapXY_points=np.array([[366015, 5916585],
                                                               [366165, 5916400]]))

        self.assertIsInstance(vals_z, np.ndarray)
        shape_exp = (2, 1, self.gA.bands)
        self.assertEqual(vals_z.shape, shape_exp)
        self.assertTrue(np.array_equal(vals_z, np.array([[5352, 5766],
                                                         [5080, 5436]]).reshape(shape_exp)))

    def test_read_pointData__multiCoords_allbands_otherPrj(self):
        X, Y = transform_any_prj(self.gA.epsg, 4326,
                                 np.array([366015.1, 366165]),
                                 np.array([5916584.9, 5916399.9]))
        vals_z = self.gA.read_pointData(mapXY_points=np.array([[X, Y]]),
                                        mapXY_points_prj=4326)

        self.assertIsInstance(vals_z, np.ndarray)
        shape_exp = (2, 1, self.gA.bands)
        self.assertEqual(vals_z.shape, shape_exp)
        self.assertTrue(np.array_equal(vals_z, np.array([[5352, 5766],
                                                         [5080, 5436]]).reshape(shape_exp)),
                        'vals_z: %s' % repr(vals_z))

    def test_read_pointData__singleCoord_offside_oneband(self):
        vals_z = self.gA.read_pointData(mapXY_points=np.array([[365900, 5916585]]),
                                        band=1,
                                        offside_val=-9999)

        self.assertIsInstance(vals_z, int)
        self.assertEqual(vals_z, -9999)

    def test_read_pointData__multiCoords_offside_oneband(self):
        xtoolarge = self.gA.box.boundsMap[1] + self.gA.xgsd
        vals_z = self.gA.read_pointData(mapXY_points=np.array([[365900, 5916585],  # x too small
                                                               [365900, 5917000],  # x too small, y to large
                                                               [xtoolarge, 5916585],  # x too large
                                                               [366015, 5916585],
                                                               [366165, 5916400]]),
                                        offside_val=-9999,
                                        band=1)

        self.assertIsInstance(vals_z, np.ndarray)
        shape_exp = (5, 1)
        self.assertEqual(vals_z.shape, shape_exp)
        self.assertTrue(np.array_equal(vals_z, np.array([[-9999],
                                                         [-9999],
                                                         [-9999],
                                                         [5766],
                                                         [5436]])))

    def test_read_pointData__multiCoords_offside_allbands(self):
        xtoolarge = self.gA.box.boundsMap[1] + self.gA.xgsd
        vals_z = self.gA.read_pointData(mapXY_points=np.array([[365900, 5916585],  # x too small
                                                               [365900, 5917000],  # x too small, y to large
                                                               [xtoolarge, 5916585],  # x too large
                                                               [366015, 5916585],
                                                               [366165, 5916400]]),
                                        offside_val=-9999)

        self.assertIsInstance(vals_z, np.ndarray)
        shape_exp = (5, 1, self.gA.bands)
        self.assertEqual(vals_z.shape, shape_exp)
        self.assertTrue(np.array_equal(vals_z, np.array([[-9999, -9999],
                                                         [-9999, -9999],
                                                         [-9999, -9999],
                                                         [5352, 5766],
                                                         [5080, 5436]]).reshape(shape_exp)))
